#!/bin/bash

CODEDIR=".."

check_existence()
{
	# check whether $(CODEDIR)/lbmat_$(MODEL) exists
	EXE=${CODEDIR}/${1}
	if [[ ! -f "${EXE}" ]]; then
		echo "${EXE} does not exist -> calling make"
		make -C .. ${1}
		if [[ ! -f ${EXE} ]]; then
			echo "${EXE} still does not exist -> check manually"
			exit
		fi
	fi
}

echo $(realpath $0)
#MODEL="d1q3_ade"
# automatic model selection based on directory name
RP=$(realpath $0)
RP1={${RP%/*}
MODEL=${RP1##*/}

echo model: ${MODEL}

check_existence lbmat_${MODEL}
check_existence simplify
check_existence export_latex
check_existence export_latex_compare

# execute lbmat_${MODEL} to create output directories
${CODEDIR}/lbmat_${MODEL}

# make symlinks
[[ ! -f simplify ]] && ln "${CODEDIR}/simplify" simplify -s
[[ ! -d latex ]] && ln "${CODEDIR}/latex" latex -s

# symbolic substitutions
for a in output*symbolic; do
	if [ ! -f "${a}/spatial_EPDE/pde_00.tex" ]; then
		../scripts/script_backsubst_searcher.py ${a}
		../export_latex ${a}
		# check whether latex files exist
		if [ ! -f "${a}/spatial_EPDE/pde_00.tex" ]; then
			echo "${a}/spatial_EPDE/pde_00.tex not found"
			exit;
		fi
	fi
done

# compare latex
../export_latex_compare output_${MODEL}

# call latex
buf_size=100000000
TEXCOMP="lualatex"
TEXFILE=supp_${MODEL}

${TEXCOMP} ${TEXFILE}
bibtex ${TEXFILE}
${TEXCOMP} ${TEXFILE}
${TEXCOMP} ${TEXFILE}

rm -f *.aux *.bbl *.blg *.stl *.log *.out *.toc

echo "script finished"
