#pragma once
#include "lbm.h"

template< typename MM>
struct D3Q27 : LBM<MM>
{
	static constexpr const char* id = "d3q27";
	void setC(matrix &C)
	{
		C = matrix(3, 27);
		int maxima_matrix[3][27] = {
		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1}
		};
		for (int i=0;i<3;i++)
		for (int j=0;j<27;j++)
			C(i,j) = (ex)(maxima_matrix[i][j]);
	}
	D3Q27() : LBM<MM>(27) {}
};

struct D3Q27_M1
{
	static constexpr const char* id = "M1";
	static void setM(matrix &M, matrix &C)
	{
		const int Q=27,D=3;
		M = matrix(Q, Q);
		for (int j=0;j<Q;j++)
		{
			// zeroth moment
			 M(0,j) = (ex)1;
			// first moments
			for (int d=0;d<D;d++) M(d+1,j) = C(d,j);
			M(4,j) = C(0,j)*C(1,j); 	// c[0]*c[1],
			M(5,j) = C(0,j)*C(2,j);	 	// c[0]*c[2],
			M(6,j) = C(1,j)*C(2,j); 	// c[1]*c[2],
			M(7,j) = C(0,j)*C(1,j)*C(2,j); 	//c[0]*c[1]*c[2],
			M(8,j) = pow(C(0,j),2); 	// c[0]^2,
			M(9,j) = pow(C(1,j),2); 	// c[1]^2,
			M(10,j) = pow(C(2,j),2);	// c[2]^2,
			M(11,j) = pow(C(0,j),2)*C(1,j); // c[0]^2*c[1],
			M(12,j) = pow(C(0,j),2)*C(2,j); // c[0]^2*c[2],
			M(13,j) = pow(C(0,j),2)*C(1,j)*C(2,j); // c[0]^2*c[1]*c[2],
			M(14,j) = pow(C(1,j),2)*C(0,j); // c[1]^2*c[0],
			M(15,j) = pow(C(1,j),2)*C(2,j); // c[1]^2*c[2],
			M(16,j) = pow(C(1,j),2)*C(0,j)*C(2,j); // c[1]^2*c[0]*c[2],
			M(17,j) = pow(C(2,j),2)*C(0,j); // c[2]^2*c[0],
			M(18,j) = pow(C(2,j),2)*C(1,j); // c[2]^2*c[1],
			M(19,j) = pow(C(2,j),2)*C(0,j)*C(1,j); // c[2]^2*c[0]*c[1],
			M(20,j) = pow(C(0,j),2)*pow(C(1,j),2); // c[0]^2*c[1]^2,
			M(21,j) = pow(C(0,j),2)*pow(C(2,j),2); // c[0]^2*c[2]^2,
			M(22,j) = pow(C(1,j),2)*pow(C(2,j),2); // c[1]^2*c[2]^2,
			M(23,j) = pow(C(0,j),2)*pow(C(1,j),2)*C(2,j); // c[0]^2*c[1]^2*c[2],
			M(24,j) = pow(C(0,j),2)*pow(C(2,j),2)*C(1,j); // c[0]^2*c[2]^2*c[1],
			M(25,j) = pow(C(1,j),2)*pow(C(2,j),2)*C(0,j); // c[1]^2*c[2]^2*c[0],
			M(26,j) = pow(C(0,j),2)*pow(C(1,j),2)*pow(C(2,j),2); // c[0]^2*c[1]^2*c[2]^2);
		}
	}

	// defined using discrete F_eq
// 	ex get_moments_eq(int q, FUN& fun)
// 	{
// 		if (q==0) return fun.rho;
// 		// using c_s^2 = 1/3
// 		if (q==1) return fun.rho*fun.vel[0];
// 		if (q==2) return fun.rho*fun.vel[1];
// 		if (q==3) return fun.rho*fun.vel[2];
// 		if (q==4) return fun.rho*(fun.vel[0]*fun.vel[1]);
// 		if (q==5) return fun.rho*(fun.vel[0]*fun.vel[2]);
// 		if (q==6) return fun.rho*(fun.vel[1]*fun.vel[2]);
// 		if (q==7) return 0;
// 		if (q==8)  return fun.rho*fun.vel[0]*fun.vel[0] + fun.rho/3;
// 		if (q==9)  return fun.rho*fun.vel[1]*fun.vel[1] + fun.rho/3;
// 		if (q==10) return fun.rho*fun.vel[2]*fun.vel[2] + fun.rho/3;
// 		if (q==11) return fun.rho*fun.vel[1]/3;
// 		if (q==12) return fun.rho*fun.vel[2]/3;
// 		if (q==13) return fun.rho*fun.vel[1]*fun.vel[2]/3;
// 		if (q==14) return fun.rho*fun.vel[0]/3;
// 		if (q==15) return fun.rho*fun.vel[2]/3;
// 		if (q==16) return fun.rho*fun.vel[0]*fun.vel[2]/3;
// 		if (q==17) return fun.rho*fun.vel[0]/3;
// 		if (q==18) return fun.rho*fun.vel[1]/3;
// 		if (q==19) return fun.rho*fun.vel[0]*fun.vel[1]/3;
// 		if (q==20) return fun.rho*(fun.vel[1]*fun.vel[1] + fun.vel[0]*fun.vel[0])/3 +fun.rho/9;
// 		if (q==21) return fun.rho*fun.vel[2]*fun.vel[2]/3 + fun.rho*fun.vel[0]*fun.vel[0]/3 + fun.rho/9;
// 		if (q==22) return fun.rho*fun.vel[2]*fun.vel[2]/3 + fun.rho*fun.vel[1]*fun.vel[1]/3 + fun.rho/9;
// 		if (q==23) return fun.rho*fun.vel[2]/9;
// 		if (q==24) return fun.rho*fun.vel[1]/9;
// 		if (q==25) return fun.rho*fun.vel[0]/9;
// 		if (q==26) return fun.rho*fun.vel[2]*fun.vel[2]/9 + fun.rho*fun.vel[1]*fun.vel[1]/9 + fun.rho*fun.vel[0]*fun.vel[0]/9 + fun.rho/27;
// 		return 0;
// 	}

	// defined using continuos F_eq
	static ex get_moments_eq(int q, FUN& fun, ex sym_cs, matrix &C)
	{
		if (q==0) return fun.rho;
		// using c_s^2 = 1/3
		if (q==1) return fun.rho*fun.vel[0];
		if (q==2) return fun.rho*fun.vel[1];
		if (q==3) return fun.rho*fun.vel[2];
		if (q==4) return fun.rho*(fun.vel[0]*fun.vel[1]);
		if (q==5) return fun.rho*(fun.vel[0]*fun.vel[2]);
		if (q==6) return fun.rho*(fun.vel[1]*fun.vel[2]);
		if (q==7) return fun.rho*fun.vel[0]*fun.vel[1]*fun.vel[2];
		if (q==8)  return fun.rho*fun.vel[0]*fun.vel[0] + fun.rho*sym_cs*sym_cs;
		if (q==9)  return fun.rho*fun.vel[1]*fun.vel[1] + fun.rho*sym_cs*sym_cs;
		if (q==10) return fun.rho*fun.vel[2]*fun.vel[2] + fun.rho*sym_cs*sym_cs;
		if (q==11) return fun.rho*(fun.vel[0]*fun.vel[0]*fun.vel[1] + fun.vel[1]*sym_cs*sym_cs);
		if (q==12) return fun.rho*(fun.vel[0]*fun.vel[0]*fun.vel[2] + fun.vel[2]*sym_cs*sym_cs);
		if (q==13) return fun.rho*(fun.vel[0]*fun.vel[0]*fun.vel[1]*fun.vel[2] + fun.vel[1]*fun.vel[2]*sym_cs*sym_cs);
		if (q==14) return fun.rho*(fun.vel[0]*fun.vel[1]*fun.vel[1] + fun.vel[0]*sym_cs*sym_cs);
		if (q==15) return fun.rho*(fun.vel[2]*fun.vel[1]*fun.vel[1] + fun.vel[2]*sym_cs*sym_cs);
		if (q==16) return fun.rho*(fun.vel[0]*fun.vel[1]*fun.vel[1]*fun.vel[2] + fun.vel[0]*fun.vel[2]*sym_cs*sym_cs);
		if (q==17) return fun.rho*(fun.vel[0]*fun.vel[2]*fun.vel[2] + fun.vel[0]*sym_cs*sym_cs);
		if (q==18) return fun.rho*(fun.vel[1]*fun.vel[2]*fun.vel[2] + fun.vel[1]*sym_cs*sym_cs);
		if (q==19) return fun.rho*(fun.vel[0]*fun.vel[1]*fun.vel[2]*fun.vel[2] + fun.vel[0]*fun.vel[1]*sym_cs*sym_cs);
		if (q==20) return fun.rho*(fun.vel[1]*fun.vel[1] + fun.vel[0]*fun.vel[0])*sym_cs*sym_cs +fun.rho*sym_cs*sym_cs*sym_cs*sym_cs + fun.rho*fun.vel[0]*fun.vel[0]*fun.vel[1]*fun.vel[1];
		if (q==21) return fun.rho*fun.vel[2]*fun.vel[2]*sym_cs*sym_cs + fun.rho*fun.vel[0]*fun.vel[0]*sym_cs*sym_cs + fun.rho*sym_cs*sym_cs*sym_cs*sym_cs + fun.rho*fun.vel[0]*fun.vel[0]*fun.vel[2]*fun.vel[2];
		if (q==22) return fun.rho*fun.vel[2]*fun.vel[2]*sym_cs*sym_cs + fun.rho*fun.vel[1]*fun.vel[1]*sym_cs*sym_cs + fun.rho*sym_cs*sym_cs*sym_cs*sym_cs + fun.rho*fun.vel[1]*fun.vel[1]*fun.vel[2]*fun.vel[2];
		if (q==23) return fun.rho*fun.vel[2]*sym_cs*sym_cs*sym_cs*sym_cs + fun.rho*fun.vel[0]*fun.vel[0]*fun.vel[1]*fun.vel[1]*fun.vel[2] + fun.rho*(fun.vel[0]*fun.vel[0]*fun.vel[2] + fun.vel[1]*fun.vel[1]*fun.vel[2])*sym_cs*sym_cs;
		if (q==24) return fun.rho*fun.vel[1]*sym_cs*sym_cs*sym_cs*sym_cs + fun.rho*fun.vel[0]*fun.vel[1]*fun.vel[1]*fun.vel[2]*fun.vel[2] + fun.rho*(fun.vel[0]*fun.vel[1]*fun.vel[1] + fun.vel[0]*fun.vel[2]*fun.vel[2])*sym_cs*sym_cs;
		if (q==25) return fun.rho*fun.vel[0]*sym_cs*sym_cs*sym_cs*sym_cs + fun.rho*fun.vel[0]*fun.vel[0]*fun.vel[1]*fun.vel[1]*fun.vel[2] + fun.rho*(fun.vel[0]*fun.vel[0]*fun.vel[2] + fun.vel[1]*fun.vel[1]*fun.vel[2])*sym_cs*sym_cs;
		if (q==26) return fun.rho*fun.vel[0]*fun.vel[0]*fun.vel[1]*fun.vel[1]*fun.vel[2]*fun.vel[2] + fun.rho*(fun.vel[0]*fun.vel[0]*fun.vel[1]*fun.vel[1] + fun.vel[0]*fun.vel[0]*fun.vel[2]*fun.vel[2] + fun.vel[1]*fun.vel[1]*fun.vel[2]*fun.vel[2])*sym_cs*sym_cs + fun.rho*(fun.vel[0]*fun.vel[0] + fun.vel[1]*fun.vel[1] + fun.vel[2]*fun.vel[2])*sym_cs*sym_cs*sym_cs*sym_cs + fun.rho*sym_cs*sym_cs*sym_cs*sym_cs*sym_cs*sym_cs;
		return 0;
	}
};

struct D3Q27_M2
{
	static constexpr const char* id = "M2";
	static void setM(matrix &M, matrix &C)
	{
		const int Q=27,D=3;
		M = matrix(Q, Q);
		for (int j=0;j<Q;j++)
		{
			M(0,j) = (ex)1;
			for (int d=0;d<D;d++) M(d+1,j) = C(d,j);
			// non-orthogonal MRT based on Geier 2015 Appendix C
			M(4,j) = C(0,j)*C(1,j); 	// c[0]*c[1], m110
			M(5,j) = C(0,j)*C(2,j);	 	// c[0]*c[2], m101
			M(6,j) = C(1,j)*C(2,j); 	// c[1]*c[2], m011
			M(7,j) = pow(C(0,j),2) - pow(C(1,j),2); 	// m200 - m020	(C.1)
			M(8,j) = pow(C(0,j),2) - pow(C(2,j),2);	 	// m200 - m002	(C.2)
			M(9,j) = pow(C(0,j),2) + pow(C(1,j),2) + pow(C(2,j),2); 	// m200 + m020 + m002	(C.3)
			M(10,j) = C(0,j)*pow(C(1,j),2) + C(0,j)*pow(C(2,j),2); 	// m120 + m102	(C.4)
			M(11,j) = pow(C(0,j),2)*C(1,j) + C(1,j)*pow(C(2,j),2); 	// m210 + m012	(C.5)
			M(12,j) = pow(C(0,j),2)*C(2,j) + pow(C(1,j),2)*C(2,j); 	// m201 + m021	(C.6)
			M(13,j) = C(0,j)*pow(C(1,j),2) - C(0,j)*pow(C(2,j),2); 	// m120 - m102	(C.7)
			M(14,j) = pow(C(0,j),2)*C(1,j) - C(1,j)*pow(C(2,j),2); 	// m210 - m012	(C.8)
			M(15,j) = pow(C(0,j),2)*C(2,j) - pow(C(1,j),2)*C(2,j); 	// m201 - m021	(C.9)
			M(16,j) = C(0,j)*C(1,j)*C(2,j); // m111	(C.10)
			M(17,j) = pow(C(0,j),2)*pow(C(1,j),2) - (ex)2*pow(C(0,j),2)*pow(C(2,j),2) + pow(C(1,j),2)*pow(C(2,j),2); // m220 - 2*m202 + m022 	(C.11)
			M(18,j) = pow(C(0,j),2)*pow(C(1,j),2) + pow(C(0,j),2)*pow(C(2,j),2) - (ex)2*pow(C(1,j),2)*pow(C(2,j),2); // m220 + m202 - 2*m022 	(C.12)
			M(19,j) = pow(C(0,j),2)*pow(C(1,j),2) + pow(C(0,j),2)*pow(C(2,j),2) + pow(C(1,j),2)*pow(C(2,j),2); // m220 + m202 + m022 	(C.13)
			M(20,j) = pow(C(0,j),2)*C(1,j)*C(2,j); // m211	(C.14)
			M(21,j) = C(0,j)*pow(C(1,j),2)*C(2,j); // m121	(C.15)
			M(22,j) = C(0,j)*C(1,j)*pow(C(2,j),2); // m112	(C.16)
			M(23,j) = pow(C(0,j),2)*pow(C(1,j),2)*C(2,j); // m221	(C.17)
			M(24,j) = pow(C(0,j),2)*C(1,j)*pow(C(2,j),2); // m212	(C.18)
			M(25,j) = C(0,j)*pow(C(1,j),2)*pow(C(2,j),2); // m122	(C.19)
			M(26,j) = pow(C(0,j),2)*pow(C(1,j),2)*pow(C(2,j),2); // m222	(C.20)
		}
	}
	static ex get_moments_eq(int q, FUN&fun, ex sym_cs, matrix &C)
	{
		const int Q=27;
		matrix mu1_eq = matrix(Q,1);
		for (int i=0;i<Q;i++)
			mu1_eq(i,0) = D3Q27_M1::get_moments_eq(i,fun,sym_cs,C);
		matrix M1,M2;
		D3Q27_M1::setM(M1,C);
		setM(M2,C);
		matrix M1i = inverse(M1);
		matrix mu2_eq = M2.mul(M1i).mul(mu1_eq);
		return mu2_eq(q,0);
	}
};


struct D3Q27_M3dev
{
	static constexpr const char* id = "M3dev";
	static void setM(matrix &M, matrix &C)
	{
		const int Q=27,D=3;
		M = matrix(Q, Q);
		for (int j=0;j<Q;j++)
		{
			// zeroth moment
			M(0,j) = (ex)1;
			// first moments
			for (int d=0;d<D;d++) M(d+1,j) = C(d,j);
			if (j>D) M(j,j) = 1; // diagonal
		}
	}
	static ex get_moments_eq(int q, FUN&fun, ex sym_cs, matrix &C)
	{
		const int Q=27;
		// workaround:
		// mu_eq = 
		matrix mu1_eq = matrix(Q,1);
		for (int i=0;i<Q;i++)
			mu1_eq(i,0) = D3Q27_M1::get_moments_eq(i,fun,sym_cs,C);
		matrix M1,M2;
		D3Q27_M1::setM(M1,C);
		setM(M2,C);
		matrix M1i = inverse(M1);
		matrix mu2_eq = M2.mul(M1i).mul(mu1_eq);
		return mu2_eq(q,0);
	}
};


template< typename MM>
struct D3Q27_SRT : D3Q27<MM>
{
	static constexpr const char* id_col = "srt";
	using LBM<MM>::Q;
	virtual void setA(matrix &A, ex& u, ex&v, ex&w)
	{
		A = matrix(Q, Q);
		for (int i=0;i<Q;i++)
		for (int j=0;j<Q;j++)
			A(i,j) = (i==j) ? (ex)LBM<MM>::sym_o : (ex)0;
	}

};

template< typename MM>
struct D3Q27_MRT1 : D3Q27<MM>
{
	static constexpr const char* id_col = "mrt1";
	virtual void setM(matrix &M, matrix &C)
	{
		D3Q27_M1::setM(M,C);
	}
	virtual void setA(matrix &A, ex& u, ex&v, ex&w)
	{
		matrix S,M;
		LBM<MM>::setS(S);
	        LBM<MM>::setM(M);
	        matrix Mi = inverse(M);
	        A = Mi.mul(S).mul(M);
	}
};

template< typename MM>
struct D3Q27_MRT2 : D3Q27_MRT1<MM>
{
	static constexpr const char* id_col = "mrt2";
	virtual void setM(matrix &M, matrix &C)
	{
		D3Q27_M2::setM(M,C);
	}
};

template< typename MM>
struct D3Q27_CLBM1 : D3Q27_MRT1<MM>
{
	static constexpr const char* id_col = "clbm1";
	virtual void setA(matrix &A, ex& u, ex&v, ex&w)
	{
		matrix S,K;
		LBM<MM>::setS(S);
		LBM<MM>::setK(K, u, v, w);
		matrix Ki = inverse(K);
		A = Ki.mul(S).mul(K);
	}
};

template< typename MM>
struct D3Q27_CLBM2 : D3Q27_CLBM1<MM>
{
	static constexpr const char* id_col = "clbm2";
	virtual void setM(matrix &M, matrix &C)
	{
		D3Q27_M2::setM(M,C);
	}

};


// cumulant by PE 2021.11.12
template< typename MM>
struct D3Q27_CuLBM1 : D3Q27<MM>
{
	static constexpr const char* id_col = "culbm1";
	static constexpr const bool is_matrix_lbm = false;
	using LBM<MM>::sym_os;
	using LBM<MM>::sym_cs;

	virtual void collision(matrix &rhs, matrix &dfs, ex& u, ex&v, ex&w, ex&rho)
	{
		// converted from lbm: lbm_col_cum.h version 2021.08.21
		// simplified way of using fractions in the CuLBM codes
		const ex no1 = (ex)1;
		const ex no3 = (ex)3;
		const ex n1o2 = 1/(ex)2;
		const ex n1o3 = 1/(ex)3;
		const ex n2o3 = 2/(ex)3;
		const ex n4o3 = 4/(ex)3;
		const ex n1o4 = 1/(ex)4;
		const ex n1o5 = 1/(ex)5;
		const ex n4o5 = 4/(ex)5;
		const ex n1o6 = 1/(ex)6;
		const ex n1o7 = 1/(ex)7;
		const ex n5o7 = 5/(ex)7;
		const ex n6o7 = 6/(ex)7;
		const ex n1o12 = 1/(ex)12;
		const ex n1o20 = 1/(ex)20;
		const ex n2o15 = 2/(ex)15;
		const ex n3o10 = 3/(ex)10;
		const ex n1o42 = 1/(ex)42;
		const ex n2o35 = 2/(ex)35;
		const ex n5o14 = 5/(ex)14;


		enum { zzz, pzz, zpz, zzp, mzz, zmz, zzm, zpp, zpm, zmp, zmm, ppz, pmz, mpz, mmz, pzp, pzm, mzp, mzm, ppp, ppm, pmp, pmm, mpp, mpm, mmp, mmm };

				//Eq 6
		ex k_mm0 = (dfs(mmp, 0)+ dfs(mmm, 0)) + dfs(mmz, 0);
		ex k_mz0 = (dfs(mzp, 0)+ dfs(mzm, 0)) + dfs(mzz, 0);
		ex k_mp0 = (dfs(mpp, 0)+ dfs(mpm, 0)) + dfs(mpz, 0);
		ex k_zm0 = (dfs(zmp, 0)+ dfs(zmm, 0)) + dfs(zmz, 0);
		ex k_zz0 = (dfs(zzp, 0)+ dfs(zzm, 0)) + dfs(zzz, 0);
		ex k_zp0 = (dfs(zpp, 0)+ dfs(zpm, 0)) + dfs(zpz, 0);
		ex k_pm0 = (dfs(pmp, 0)+ dfs(pmm, 0)) + dfs(pmz, 0);
		ex k_pz0 = (dfs(pzp, 0)+ dfs(pzm, 0)) + dfs(pzz, 0);
		ex k_pp0 = (dfs(ppp, 0)+ dfs(ppm, 0)) + dfs(ppz, 0);

		//Eq 7
		ex k_mm1 = (dfs(mmp, 0)- dfs(mmm, 0)) - w * k_mm0;
		ex k_mz1 = (dfs(mzp, 0)- dfs(mzm, 0)) - w * k_mz0;
		ex k_mp1 = (dfs(mpp, 0)- dfs(mpm, 0)) - w * k_mp0;
		ex k_zm1 = (dfs(zmp, 0)- dfs(zmm, 0)) - w * k_zm0;
		ex k_zz1 = (dfs(zzp, 0)- dfs(zzm, 0)) - w * k_zz0;
		ex k_zp1 = (dfs(zpp, 0)- dfs(zpm, 0)) - w * k_zp0;
		ex k_pm1 = (dfs(pmp, 0)- dfs(pmm, 0)) - w * k_pm0;
		ex k_pz1 = (dfs(pzp, 0)- dfs(pzm, 0)) - w * k_pz0;
		ex k_pp1 = (dfs(ppp, 0)- dfs(ppm, 0)) - w * k_pp0;

		//Eq 8
		ex k_mm2 = (dfs(mmp, 0)+ dfs(mmm, 0)) - 2*w*(dfs(mmp, 0)- dfs(mmm, 0)) + w*w* k_mm0;
		ex k_mz2 = (dfs(mzp, 0)+ dfs(mzm, 0)) - 2*w*(dfs(mzp, 0)- dfs(mzm, 0)) + w*w* k_mz0;
		ex k_mp2 = (dfs(mpp, 0)+ dfs(mpm, 0)) - 2*w*(dfs(mpp, 0)- dfs(mpm, 0)) + w*w* k_mp0;
		ex k_zm2 = (dfs(zmp, 0)+ dfs(zmm, 0)) - 2*w*(dfs(zmp, 0)- dfs(zmm, 0)) + w*w* k_zm0;
		ex k_zz2 = (dfs(zzp, 0)+ dfs(zzm, 0)) - 2*w*(dfs(zzp, 0)- dfs(zzm, 0)) + w*w* k_zz0;
		ex k_zp2 = (dfs(zpp, 0)+ dfs(zpm, 0)) - 2*w*(dfs(zpp, 0)- dfs(zpm, 0)) + w*w* k_zp0;
		ex k_pm2 = (dfs(pmp, 0)+ dfs(pmm, 0)) - 2*w*(dfs(pmp, 0)- dfs(pmm, 0)) + w*w* k_pm0;
		ex k_pz2 = (dfs(pzp, 0)+ dfs(pzm, 0)) - 2*w*(dfs(pzp, 0)- dfs(pzm, 0)) + w*w* k_pz0;
		ex k_pp2 = (dfs(ppp, 0)+ dfs(ppm, 0)) - 2*w*(dfs(ppp, 0)- dfs(ppm, 0)) + w*w* k_pp0;

		//Eq 9
		ex k_m00 = (k_mp0 + k_mm0) + k_mz0;
		ex k_z00 = (k_zp0 + k_zm0) + k_zz0;
		ex k_p00 = (k_pp0 + k_pm0) + k_pz0;
		ex k_m01 = (k_mp1 + k_mm1) + k_mz1;
		ex k_z01 = (k_zp1 + k_zm1) + k_zz1;
		ex k_p01 = (k_pp1 + k_pm1) + k_pz1;
		ex k_m02 = (k_mp2 + k_mm2) + k_mz2;
		ex k_z02 = (k_zp2 + k_zm2) + k_zz2;
		ex k_p02 = (k_pp2 + k_pm2) + k_pz2;

		//Eq 10
		ex k_m10 = (k_mp0 - k_mm0) - v * k_m00;
		ex k_z10 = (k_zp0 - k_zm0) - v * k_z00;
		ex k_p10 = (k_pp0 - k_pm0) - v * k_p00;
		ex k_m11 = (k_mp1 - k_mm1) - v * k_m01;
		ex k_z11 = (k_zp1 - k_zm1) - v * k_z01;
		ex k_p11 = (k_pp1 - k_pm1) - v * k_p01;
		ex k_m12 = (k_mp2 - k_mm2) - v * k_m02;
		ex k_z12 = (k_zp2 - k_zm2) - v * k_z02;
		ex k_p12 = (k_pp2 - k_pm2) - v * k_p02;

		//Eq 11
		ex k_m20 = (k_mp0 + k_mm0) - 2*v* (k_mp0 - k_mm0) + v*v * k_m00;
		ex k_z20 = (k_zp0 + k_zm0) - 2*v* (k_zp0 - k_zm0) + v*v * k_z00;
		ex k_p20 = (k_pp0 + k_pm0) - 2*v* (k_pp0 - k_pm0) + v*v * k_p00;
		ex k_m21 = (k_mp1 + k_mm1) - 2*v* (k_mp1 - k_mm1) + v*v * k_m01;
		ex k_z21 = (k_zp1 + k_zm1) - 2*v* (k_zp1 - k_zm1) + v*v * k_z01;
		ex k_p21 = (k_pp1 + k_pm1) - 2*v* (k_pp1 - k_pm1) + v*v * k_p01;
		ex k_m22 = (k_mp2 + k_mm2) - 2*v* (k_mp2 - k_mm2) + v*v * k_m02;
		ex k_z22 = (k_zp2 + k_zm2) - 2*v* (k_zp2 - k_zm2) + v*v * k_z02;
		ex k_p22 = (k_pp2 + k_pm2) - 2*v* (k_pp2 - k_pm2) + v*v * k_p02;

		//Eq 12
		ex k_000 = (k_p00 + k_m00) + k_z00;
		ex k_001 = (k_p01 + k_m01) + k_z01;
		ex k_002 = (k_p02 + k_m02) + k_z02;
		ex k_010 = (k_p10 + k_m10) + k_z10;
		ex k_011 = (k_p11 + k_m11) + k_z11;
		ex k_012 = (k_p12 + k_m12) + k_z12;
		ex k_020 = (k_p20 + k_m20) + k_z20;
		ex k_021 = (k_p21 + k_m21) + k_z21;
		ex k_022 = (k_p22 + k_m22) + k_z22;

		//Eq 13
		ex k_100 = (k_p00 - k_m00) - u * k_000 ;
		ex k_101 = (k_p01 - k_m01) - u * k_001 ;
		ex k_102 = (k_p02 - k_m02) - u * k_002 ;
		ex k_110 = (k_p10 - k_m10) - u * k_010 ;
		ex k_111 = (k_p11 - k_m11) - u * k_011 ;
		ex k_112 = (k_p12 - k_m12) - u * k_012 ;
		ex k_120 = (k_p20 - k_m20) - u * k_020 ;
		ex k_121 = (k_p21 - k_m21) - u * k_021 ;
		ex k_122 = (k_p22 - k_m22) - u * k_022 ;

		//Eq 14
		ex k_200 = (k_p00 + k_m00) - 2*u* (k_p00 - k_m00) + u*u * k_000 ;
		ex k_201 = (k_p01 + k_m01) - 2*u* (k_p01 - k_m01) + u*u * k_001 ;
		ex k_202 = (k_p02 + k_m02) - 2*u* (k_p02 - k_m02) + u*u * k_002 ;
		ex k_210 = (k_p10 + k_m10) - 2*u* (k_p10 - k_m10) + u*u * k_010 ;
		ex k_211 = (k_p11 + k_m11) - 2*u* (k_p11 - k_m11) + u*u * k_011 ;
		ex k_212 = (k_p12 + k_m12) - 2*u* (k_p12 - k_m12) + u*u * k_012 ;
		ex k_220 = (k_p20 + k_m20) - 2*u* (k_p20 - k_m20) + u*u * k_020 ;
		ex k_221 = (k_p21 + k_m21) - 2*u* (k_p21 - k_m21) + u*u * k_021 ;
		ex k_222 = (k_p22 + k_m22) - 2*u* (k_p22 - k_m22) + u*u * k_022 ;

		//Eq G2015(51)
		ex C_211 = k_211 - (k_200*k_011 + 2*k_101*k_110)/rho;
		ex C_121 = k_121 - (k_020*k_101 + 2*k_110*k_011)/rho;
		ex C_112 = k_112 - (k_002*k_110 + 2*k_011*k_101)/rho;

		//Eq G2015(52)
		ex C_220 = k_220- (k_020*k_200 + 2*k_110*k_110)/rho;
		ex C_022 = k_022- (k_002*k_020 + 2*k_011*k_011)/rho;
		ex C_202 = k_202- (k_200*k_002 + 2*k_101*k_101)/rho;

		//Eq G2015(53)
		ex C_122 = k_122- (k_020*k_102 + k_002*k_120 + 4*k_011*k_111 + 2*( k_110*k_012 + k_101*k_021))/rho;
		ex C_212 = k_212- (k_002*k_210 + k_200*k_012 + 4*k_101*k_111 + 2*( k_011*k_201 + k_110*k_102))/rho;
		ex C_221 = k_221- (k_200*k_021 + k_020*k_201 + 4*k_110*k_111 + 2*( k_101*k_120 + k_011*k_210))/rho;
		//Eq G2015(54)
		ex C_222 = k_222 - ( 4*k_111*k_111 + k_200*k_022 + k_020*k_202 + k_002*k_220 + 4*(k_011*k_211 + k_101*k_121 + k_110*k_112) + 2*(k_120*k_102 + k_210*k_012 + k_201*k_021))/rho +(16*k_110*k_101*k_011 + 4*(k_101*k_101*k_020 + k_011*k_011*k_200 + k_110*k_110*k_002)+2*k_200*k_020*k_002)/rho/rho;
		// gen1nowell.php END

// 		// relaxation definition
// 		ex omega1 = 1/(no3*KS.lbmViscosity+(ex)1/(ex)2); // shear viscosity
// 		ex omega2 = 1;///(no3*KS.lbmViscosity*2 + (ex)1/(ex)2); // bulkViscosity > Viscosity ... test: bulkViscosity = 2/3 shearViscosity
// 
// 		ex omega3 = 1;
// 		ex omega4 = 1;
// 		ex omega5 = 1;
// 		ex omega6 = 1;
// 		ex omega7 = 1;
// 		ex omega8 = 1;
// 		ex omega9 = 1;
// 		ex omega10 = 1;

 		ex omega1 = sym_os[0];
 		ex omega2 = sym_os[1];
 		ex omega3 = sym_os[2];
 		ex omega4 = sym_os[3];
 		ex omega5 = sym_os[4];
 		ex omega6 = sym_os[5];
 		ex omega7 = sym_os[6];
 		ex omega8 = sym_os[7];
 		ex omega9 = sym_os[8];
 		ex omega10 = sym_os[9];
        ex omega11 = sym_os[10];
        ex omega12 = sym_os[11];
        ex omega13 = sym_os[12];
        ex omega14 = sym_os[13];
        ex omega15 = sym_os[14];
        ex omega16 = sym_os[15];
        ex omega17 = sym_os[16];
        ex omega18 = sym_os[17];
        ex omega19 = sym_os[18];
        ex omega20 = sym_os[19];
        ex omega21 = sym_os[20];
        ex omega22 = sym_os[21];
        ex omega23 = sym_os[22];

		// extra parameters

		// actual collision step: note: ks = Cs for these indexes
		ex Cs_110 = (1 - omega1)*k_110;
		ex Cs_101 = (1 - omega2)*k_101;
		ex Cs_011 = (1 - omega3)*k_011;

		// plus their combination: Eq 30 - 32

		// see rovnice2.mw
		ex Cs_200 = (1 - omega4)*k_200 + omega4*sym_cs*sym_cs*k_000;
		ex Cs_020 = (1 - omega5)*k_020 + omega5*sym_cs*sym_cs*k_000;
		ex Cs_002 = (1 - omega6)*k_002 + omega6*sym_cs*sym_cs*k_000;

		// Eqs 36-41: see vzorce2.mw
		ex Cs_120 = (1 - omega7)*k_120;
		ex Cs_102 = (1 - omega8)*k_102;
		ex Cs_210 = (1 - omega9)*k_210;
		ex Cs_012 = (1 - omega10)*k_012;
		ex Cs_021 = (1 - omega11)*k_021;
		ex Cs_201 = (1 - omega12)*k_201;
		// Eq 42
		ex Cs_111 = (1-omega13)*k_111;

		// see rovnice2.mw !!! beware
		ex Cs_220 = (1 - omega14)*C_220;
		ex Cs_202 = (1 - omega15)*C_202;
		ex Cs_022 = (1 - omega16)*C_022;
		// Eq 46-48
		ex Cs_211 = (1-omega17)*C_211;
		ex Cs_121 = (1-omega18)*C_121;
		ex Cs_112 = (1-omega19)*C_112;
		// Eqs 49-52
		ex Cs_221 = (1-omega20)*C_221;
		ex Cs_212 = (1-omega21)*C_212;
		ex Cs_122 = (1-omega22)*C_122;
		ex Cs_222 = (1-omega23)*C_222;

// remark: collision step is the same as in well-CUM

		// 3.4 Backward cumulant transformation
		// gen2nowell.php BEGIN
		//Eq. G2015(81)
		ex ks_211 = Cs_211+ (Cs_200*Cs_011 + 2*Cs_101*Cs_110)/rho;
		ex ks_121 = Cs_121+ (Cs_020*Cs_101 + 2*Cs_110*Cs_011)/rho;
		ex ks_112 = Cs_112+ (Cs_002*Cs_110 + 2*Cs_011*Cs_101)/rho;

		//Eq. G2015(82)
		ex ks_220 = Cs_220+ (Cs_020*Cs_200 + 2*Cs_110*Cs_110)/rho;
		ex ks_022 = Cs_022+ (Cs_002*Cs_020 + 2*Cs_011*Cs_011)/rho;
		ex ks_202 = Cs_202+ (Cs_200*Cs_002 + 2*Cs_101*Cs_101)/rho;

		//Eq. G2015(83)
		ex ks_122 = Cs_122+ (Cs_020*Cs_102 + Cs_002*Cs_120+ 4*Cs_011*Cs_111+ 2*(Cs_110*Cs_012+Cs_101*Cs_021))/rho;
		ex ks_212 = Cs_212+ (Cs_002*Cs_210 + Cs_200*Cs_012+ 4*Cs_101*Cs_111+ 2*(Cs_011*Cs_201+Cs_110*Cs_102))/rho;
		ex ks_221 = Cs_221+ (Cs_200*Cs_021 + Cs_020*Cs_201+ 4*Cs_110*Cs_111+ 2*(Cs_101*Cs_120+Cs_011*Cs_210))/rho;
		// gen2nowell.php END
		
		// Eq. G2015(84)
		ex ks_222 = Cs_222 + (4*Cs_111*Cs_111 + Cs_200*ks_022 + Cs_020*ks_202 + Cs_002*ks_220 + 4*(Cs_011*ks_211 + Cs_101*ks_121 + Cs_110*ks_112)
					+ 2*(Cs_120*Cs_102 + Cs_210*Cs_012 + Cs_201*Cs_021))/rho
					- (16*Cs_110*Cs_101*Cs_011 + 4*(Cs_101*Cs_101*Cs_020 + Cs_011*Cs_011*Cs_200 + Cs_110*Cs_110*Cs_002) + 2*Cs_200*Cs_020*Cs_002)/rho/rho;


		// backward central moment transformation
		ex ks_000 =  k_000;
		// Geier 2017: forcing scheme ---> zde se musi dat (-1) pro forcing ... ale s nulovou silou a - zde to nefunguje ... 
		ex Cs_100 =  -k_100;
		ex Cs_010 =  -k_010;
		ex Cs_001 =  -k_001;
		
		// gen3nowell.php BEGIN
		//Eq G2015(88)
		ex ks_z00 = ks_000*(1-u*u) - 2*u*Cs_100 - Cs_200;
		ex ks_z01 = Cs_001*(1-u*u) - 2*u*Cs_101 - Cs_201;
		ex ks_z02 = Cs_002*(1-u*u) - 2*u*Cs_102 - ks_202;
		ex ks_z10 = Cs_010*(1-u*u) - 2*u*Cs_110 - Cs_210;
		ex ks_z11 = Cs_011*(1-u*u) - 2*u*Cs_111 - ks_211;
		ex ks_z12 = Cs_012*(1-u*u) - 2*u*ks_112 - ks_212;
		ex ks_z20 = Cs_020*(1-u*u) - 2*u*Cs_120 - ks_220;
		ex ks_z21 = Cs_021*(1-u*u) - 2*u*ks_121 - ks_221;
		ex ks_z22 = ks_022*(1-u*u) - 2*u*ks_122 - ks_222;

		//Eq  G2015(89)
		ex ks_m00 = (ks_000*(u*u - u) + Cs_100*(2*u - 1) + Cs_200)*n1o2;
		ex ks_m01 = (Cs_001*(u*u - u) + Cs_101*(2*u - 1) + Cs_201)*n1o2;
		ex ks_m02 = (Cs_002*(u*u - u) + Cs_102*(2*u - 1) + ks_202)*n1o2;
		ex ks_m10 = (Cs_010*(u*u - u) + Cs_110*(2*u - 1) + Cs_210)*n1o2;
		ex ks_m11 = (Cs_011*(u*u - u) + Cs_111*(2*u - 1) + ks_211)*n1o2;
		ex ks_m12 = (Cs_012*(u*u - u) + ks_112*(2*u - 1) + ks_212)*n1o2;
		ex ks_m20 = (Cs_020*(u*u - u) + Cs_120*(2*u - 1) + ks_220)*n1o2;
		ex ks_m21 = (Cs_021*(u*u - u) + ks_121*(2*u - 1) + ks_221)*n1o2;
		ex ks_m22 = (ks_022*(u*u - u) + ks_122*(2*u - 1) + ks_222)*n1o2;

		//Eq  G2015(90)
		ex ks_p00 = (ks_000*(u*u + u) + Cs_100*(2*u + 1) + Cs_200)*n1o2;
		ex ks_p01 = (Cs_001*(u*u + u) + Cs_101*(2*u + 1) + Cs_201)*n1o2;
		ex ks_p02 = (Cs_002*(u*u + u) + Cs_102*(2*u + 1) + ks_202)*n1o2;
		ex ks_p10 = (Cs_010*(u*u + u) + Cs_110*(2*u + 1) + Cs_210)*n1o2;
		ex ks_p11 = (Cs_011*(u*u + u) + Cs_111*(2*u + 1) + ks_211)*n1o2;
		ex ks_p12 = (Cs_012*(u*u + u) + ks_112*(2*u + 1) + ks_212)*n1o2;
		ex ks_p20 = (Cs_020*(u*u + u) + Cs_120*(2*u + 1) + ks_220)*n1o2;
		ex ks_p21 = (Cs_021*(u*u + u) + ks_121*(2*u + 1) + ks_221)*n1o2;
		ex ks_p22 = (ks_022*(u*u + u) + ks_122*(2*u + 1) + ks_222)*n1o2;

		//Eq G2015(91)
		ex ks_mz0 = ks_m00*(1 - v*v) - 2*v*ks_m10 - ks_m20;
		ex ks_mz1 = ks_m01*(1 - v*v) - 2*v*ks_m11 - ks_m21;
		ex ks_mz2 = ks_m02*(1 - v*v) - 2*v*ks_m12 - ks_m22;
		ex ks_zz0 = ks_z00*(1 - v*v) - 2*v*ks_z10 - ks_z20;
		ex ks_zz1 = ks_z01*(1 - v*v) - 2*v*ks_z11 - ks_z21;
		ex ks_zz2 = ks_z02*(1 - v*v) - 2*v*ks_z12 - ks_z22;
		ex ks_pz0 = ks_p00*(1 - v*v) - 2*v*ks_p10 - ks_p20;
		ex ks_pz1 = ks_p01*(1 - v*v) - 2*v*ks_p11 - ks_p21;
		ex ks_pz2 = ks_p02*(1 - v*v) - 2*v*ks_p12 - ks_p22;

		//Eq  G2015(92)
		ex ks_mm0 = (ks_m00*(v*v - v) + ks_m10*(2*v - 1) + ks_m20)*n1o2;
		ex ks_mm1 = (ks_m01*(v*v - v) + ks_m11*(2*v - 1) + ks_m21)*n1o2;
		ex ks_mm2 = (ks_m02*(v*v - v) + ks_m12*(2*v - 1) + ks_m22)*n1o2;
		ex ks_zm0 = (ks_z00*(v*v - v) + ks_z10*(2*v - 1) + ks_z20)*n1o2;
		ex ks_zm1 = (ks_z01*(v*v - v) + ks_z11*(2*v - 1) + ks_z21)*n1o2;
		ex ks_zm2 = (ks_z02*(v*v - v) + ks_z12*(2*v - 1) + ks_z22)*n1o2;
		ex ks_pm0 = (ks_p00*(v*v - v) + ks_p10*(2*v - 1) + ks_p20)*n1o2;
		ex ks_pm1 = (ks_p01*(v*v - v) + ks_p11*(2*v - 1) + ks_p21)*n1o2;
		ex ks_pm2 = (ks_p02*(v*v - v) + ks_p12*(2*v - 1) + ks_p22)*n1o2;

		//Eq G2015(93)
		ex ks_mp0 = (ks_m00*(v*v + v) + ks_m10*(2*v + 1) + ks_m20)*n1o2;
		ex ks_mp1 = (ks_m01*(v*v + v) + ks_m11*(2*v + 1) + ks_m21)*n1o2;
		ex ks_mp2 = (ks_m02*(v*v + v) + ks_m12*(2*v + 1) + ks_m22)*n1o2;
		ex ks_zp0 = (ks_z00*(v*v + v) + ks_z10*(2*v + 1) + ks_z20)*n1o2;
		ex ks_zp1 = (ks_z01*(v*v + v) + ks_z11*(2*v + 1) + ks_z21)*n1o2;
		ex ks_zp2 = (ks_z02*(v*v + v) + ks_z12*(2*v + 1) + ks_z22)*n1o2;
		ex ks_pp0 = (ks_p00*(v*v + v) + ks_p10*(2*v + 1) + ks_p20)*n1o2;
		ex ks_pp1 = (ks_p01*(v*v + v) + ks_p11*(2*v + 1) + ks_p21)*n1o2;
		ex ks_pp2 = (ks_p02*(v*v + v) + ks_p12*(2*v + 1) + ks_p22)*n1o2;

		//Eq G2015(94)
		rhs(mmz, 0) =  ks_mm0*(1-w*w) - 2*w*ks_mm1 - ks_mm2;
		rhs(mzz, 0) =  ks_mz0*(1-w*w) - 2*w*ks_mz1 - ks_mz2;
		rhs(mpz, 0) =  ks_mp0*(1-w*w) - 2*w*ks_mp1 - ks_mp2;
		rhs(zmz, 0) =  ks_zm0*(1-w*w) - 2*w*ks_zm1 - ks_zm2;
		rhs(zzz, 0) =  ks_zz0*(1-w*w) - 2*w*ks_zz1 - ks_zz2;
		rhs(zpz, 0) =  ks_zp0*(1-w*w) - 2*w*ks_zp1 - ks_zp2;
		rhs(pmz, 0) =  ks_pm0*(1-w*w) - 2*w*ks_pm1 - ks_pm2;
		rhs(pzz, 0) =  ks_pz0*(1-w*w) - 2*w*ks_pz1 - ks_pz2;
		rhs(ppz, 0) =  ks_pp0*(1-w*w) - 2*w*ks_pp1 - ks_pp2;

		//Eq  G2015(95)
		rhs(mmm, 0) =  (ks_mm0*(w*w-w) + ks_mm1*(2*w - 1) + ks_mm2)*n1o2;
		rhs(mzm, 0) =  (ks_mz0*(w*w-w) + ks_mz1*(2*w - 1) + ks_mz2)*n1o2;
		rhs(mpm, 0) =  (ks_mp0*(w*w-w) + ks_mp1*(2*w - 1) + ks_mp2)*n1o2;
		rhs(zmm, 0) =  (ks_zm0*(w*w-w) + ks_zm1*(2*w - 1) + ks_zm2)*n1o2;
		rhs(zzm, 0) =  (ks_zz0*(w*w-w) + ks_zz1*(2*w - 1) + ks_zz2)*n1o2;
		rhs(zpm, 0) =  (ks_zp0*(w*w-w) + ks_zp1*(2*w - 1) + ks_zp2)*n1o2;
		rhs(pmm, 0) =  (ks_pm0*(w*w-w) + ks_pm1*(2*w - 1) + ks_pm2)*n1o2;
		rhs(pzm, 0) =  (ks_pz0*(w*w-w) + ks_pz1*(2*w - 1) + ks_pz2)*n1o2;
		rhs(ppm, 0) =  (ks_pp0*(w*w-w) + ks_pp1*(2*w - 1) + ks_pp2)*n1o2;

		//Eq  G2015(96)
		rhs(mmp, 0) =  (ks_mm0*(w*w+w) + ks_mm1*(2*w + 1) + ks_mm2)*n1o2;
		rhs(mzp, 0) =  (ks_mz0*(w*w+w) + ks_mz1*(2*w + 1) + ks_mz2)*n1o2;
		rhs(mpp, 0) =  (ks_mp0*(w*w+w) + ks_mp1*(2*w + 1) + ks_mp2)*n1o2;
		rhs(zmp, 0) =  (ks_zm0*(w*w+w) + ks_zm1*(2*w + 1) + ks_zm2)*n1o2;
		rhs(zzp, 0) =  (ks_zz0*(w*w+w) + ks_zz1*(2*w + 1) + ks_zz2)*n1o2;
		rhs(zpp, 0) =  (ks_zp0*(w*w+w) + ks_zp1*(2*w + 1) + ks_zp2)*n1o2;
		rhs(pmp, 0) =  (ks_pm0*(w*w+w) + ks_pm1*(2*w + 1) + ks_pm2)*n1o2;
		rhs(pzp, 0) =  (ks_pz0*(w*w+w) + ks_pz1*(2*w + 1) + ks_pz2)*n1o2;
		rhs(ppp, 0) =  (ks_pp0*(w*w+w) + ks_pp1*(2*w + 1) + ks_pp2)*n1o2;
	}
};


// culbm based on Geier with summation of cumulants
template< typename MM>
struct D3Q27_CuLBM2 : D3Q27<MM>
{
	static constexpr const char* id_col = "culbm2";
	static constexpr const bool is_matrix_lbm = false;
	using LBM<MM>::sym_os;
	using LBM<MM>::sym_cs;

	virtual void collision(matrix &rhs, matrix &dfs, ex& u, ex&v, ex&w, ex&rho)
	{
		// converted from lbm: lbm_col_cum.h version 2021.08.21
		// simplified way of using fractions in the CuLBM codes
		const ex no1 = (ex)1;
		const ex no3 = (ex)3;
		const ex n1o2 = 1/(ex)2;
		const ex n1o3 = 1/(ex)3;
		const ex n2o3 = 2/(ex)3;
		const ex n4o3 = 4/(ex)3;
		const ex n1o4 = 1/(ex)4;
		const ex n1o5 = 1/(ex)5;
		const ex n4o5 = 4/(ex)5;
		const ex n1o6 = 1/(ex)6;
		const ex n1o7 = 1/(ex)7;
		const ex n5o7 = 5/(ex)7;
		const ex n6o7 = 6/(ex)7;
		const ex n1o12 = 1/(ex)12;
		const ex n1o20 = 1/(ex)20;
		const ex n2o15 = 2/(ex)15;
		const ex n3o10 = 3/(ex)10;
		const ex n1o42 = 1/(ex)42;
		const ex n2o35 = 2/(ex)35;
		const ex n5o14 = 5/(ex)14;
		// following the defintion of C for D3Q27
		// C = matrix(D, Q);
		//		int maxima_matrix[3][27] = {
		//		{0,	1,	0,	0,	-1,	0,	0,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1,	1,	1,	1,	1,	-1,	-1,	-1,	-1},
		//		{0,	0,	1,	0,	0,	-1,	0,	1,	1,	-1,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	1,	-1,	-1,	1,	1,	-1,	-1},
		//		{0,	0,	0,	1,	0,	0,	-1,	1,	-1,	1,	-1,	0,	0,	0,	0,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1,	1,	-1}
		//		};
		enum { zzz, pzz, zpz, zzp, mzz, zmz, zzm, zpp, zpm, zmp, zmm, ppz, pmz, mpz, mmz, pzp, pzm, mzp, mzm, ppp, ppm, pmp, pmm, mpp, mpm, mmp, mmm };

			//Eq 6
		ex k_mm0 = (dfs(mmp, 0)+ dfs(mmm, 0)) + dfs(mmz, 0);
		ex k_mz0 = (dfs(mzp, 0)+ dfs(mzm, 0)) + dfs(mzz, 0);
		ex k_mp0 = (dfs(mpp, 0)+ dfs(mpm, 0)) + dfs(mpz, 0);
		ex k_zm0 = (dfs(zmp, 0)+ dfs(zmm, 0)) + dfs(zmz, 0);
		ex k_zz0 = (dfs(zzp, 0)+ dfs(zzm, 0)) + dfs(zzz, 0);
		ex k_zp0 = (dfs(zpp, 0)+ dfs(zpm, 0)) + dfs(zpz, 0);
		ex k_pm0 = (dfs(pmp, 0)+ dfs(pmm, 0)) + dfs(pmz, 0);
		ex k_pz0 = (dfs(pzp, 0)+ dfs(pzm, 0)) + dfs(pzz, 0);
		ex k_pp0 = (dfs(ppp, 0)+ dfs(ppm, 0)) + dfs(ppz, 0);

		//Eq 7
		ex k_mm1 = (dfs(mmp, 0)- dfs(mmm, 0)) - w * k_mm0;
		ex k_mz1 = (dfs(mzp, 0)- dfs(mzm, 0)) - w * k_mz0;
		ex k_mp1 = (dfs(mpp, 0)- dfs(mpm, 0)) - w * k_mp0;
		ex k_zm1 = (dfs(zmp, 0)- dfs(zmm, 0)) - w * k_zm0;
		ex k_zz1 = (dfs(zzp, 0)- dfs(zzm, 0)) - w * k_zz0;
		ex k_zp1 = (dfs(zpp, 0)- dfs(zpm, 0)) - w * k_zp0;
		ex k_pm1 = (dfs(pmp, 0)- dfs(pmm, 0)) - w * k_pm0;
		ex k_pz1 = (dfs(pzp, 0)- dfs(pzm, 0)) - w * k_pz0;
		ex k_pp1 = (dfs(ppp, 0)- dfs(ppm, 0)) - w * k_pp0;

		//Eq 8
		ex k_mm2 = (dfs(mmp, 0)+ dfs(mmm, 0)) - 2*w*(dfs(mmp, 0)- dfs(mmm, 0)) + w*w* k_mm0;
		ex k_mz2 = (dfs(mzp, 0)+ dfs(mzm, 0)) - 2*w*(dfs(mzp, 0)- dfs(mzm, 0)) + w*w* k_mz0;
		ex k_mp2 = (dfs(mpp, 0)+ dfs(mpm, 0)) - 2*w*(dfs(mpp, 0)- dfs(mpm, 0)) + w*w* k_mp0;
		ex k_zm2 = (dfs(zmp, 0)+ dfs(zmm, 0)) - 2*w*(dfs(zmp, 0)- dfs(zmm, 0)) + w*w* k_zm0;
		ex k_zz2 = (dfs(zzp, 0)+ dfs(zzm, 0)) - 2*w*(dfs(zzp, 0)- dfs(zzm, 0)) + w*w* k_zz0;
		ex k_zp2 = (dfs(zpp, 0)+ dfs(zpm, 0)) - 2*w*(dfs(zpp, 0)- dfs(zpm, 0)) + w*w* k_zp0;
		ex k_pm2 = (dfs(pmp, 0)+ dfs(pmm, 0)) - 2*w*(dfs(pmp, 0)- dfs(pmm, 0)) + w*w* k_pm0;
		ex k_pz2 = (dfs(pzp, 0)+ dfs(pzm, 0)) - 2*w*(dfs(pzp, 0)- dfs(pzm, 0)) + w*w* k_pz0;
		ex k_pp2 = (dfs(ppp, 0)+ dfs(ppm, 0)) - 2*w*(dfs(ppp, 0)- dfs(ppm, 0)) + w*w* k_pp0;

		//Eq 9
		ex k_m00 = (k_mp0 + k_mm0) + k_mz0;
		ex k_z00 = (k_zp0 + k_zm0) + k_zz0;
		ex k_p00 = (k_pp0 + k_pm0) + k_pz0;
		ex k_m01 = (k_mp1 + k_mm1) + k_mz1;
		ex k_z01 = (k_zp1 + k_zm1) + k_zz1;
		ex k_p01 = (k_pp1 + k_pm1) + k_pz1;
		ex k_m02 = (k_mp2 + k_mm2) + k_mz2;
		ex k_z02 = (k_zp2 + k_zm2) + k_zz2;
		ex k_p02 = (k_pp2 + k_pm2) + k_pz2;

		//Eq 10
		ex k_m10 = (k_mp0 - k_mm0) - v * k_m00;
		ex k_z10 = (k_zp0 - k_zm0) - v * k_z00;
		ex k_p10 = (k_pp0 - k_pm0) - v * k_p00;
		ex k_m11 = (k_mp1 - k_mm1) - v * k_m01;
		ex k_z11 = (k_zp1 - k_zm1) - v * k_z01;
		ex k_p11 = (k_pp1 - k_pm1) - v * k_p01;
		ex k_m12 = (k_mp2 - k_mm2) - v * k_m02;
		ex k_z12 = (k_zp2 - k_zm2) - v * k_z02;
		ex k_p12 = (k_pp2 - k_pm2) - v * k_p02;

		//Eq 11
		ex k_m20 = (k_mp0 + k_mm0) - 2*v* (k_mp0 - k_mm0) + v*v * k_m00;
		ex k_z20 = (k_zp0 + k_zm0) - 2*v* (k_zp0 - k_zm0) + v*v * k_z00;
		ex k_p20 = (k_pp0 + k_pm0) - 2*v* (k_pp0 - k_pm0) + v*v * k_p00;
		ex k_m21 = (k_mp1 + k_mm1) - 2*v* (k_mp1 - k_mm1) + v*v * k_m01;
		ex k_z21 = (k_zp1 + k_zm1) - 2*v* (k_zp1 - k_zm1) + v*v * k_z01;
		ex k_p21 = (k_pp1 + k_pm1) - 2*v* (k_pp1 - k_pm1) + v*v * k_p01;
		ex k_m22 = (k_mp2 + k_mm2) - 2*v* (k_mp2 - k_mm2) + v*v * k_m02;
		ex k_z22 = (k_zp2 + k_zm2) - 2*v* (k_zp2 - k_zm2) + v*v * k_z02;
		ex k_p22 = (k_pp2 + k_pm2) - 2*v* (k_pp2 - k_pm2) + v*v * k_p02;

		//Eq 12
		ex k_000 = (k_p00 + k_m00) + k_z00;
		ex k_001 = (k_p01 + k_m01) + k_z01;
		ex k_002 = (k_p02 + k_m02) + k_z02;
		ex k_010 = (k_p10 + k_m10) + k_z10;
		ex k_011 = (k_p11 + k_m11) + k_z11;
		ex k_012 = (k_p12 + k_m12) + k_z12;
		ex k_020 = (k_p20 + k_m20) + k_z20;
		ex k_021 = (k_p21 + k_m21) + k_z21;
		ex k_022 = (k_p22 + k_m22) + k_z22;

		//Eq 13
		ex k_100 = (k_p00 - k_m00) - u * k_000 ;
		ex k_101 = (k_p01 - k_m01) - u * k_001 ;
		ex k_102 = (k_p02 - k_m02) - u * k_002 ;
		ex k_110 = (k_p10 - k_m10) - u * k_010 ;
		ex k_111 = (k_p11 - k_m11) - u * k_011 ;
		ex k_112 = (k_p12 - k_m12) - u * k_012 ;
		ex k_120 = (k_p20 - k_m20) - u * k_020 ;
		ex k_121 = (k_p21 - k_m21) - u * k_021 ;
		ex k_122 = (k_p22 - k_m22) - u * k_022 ;

		//Eq 14
		ex k_200 = (k_p00 + k_m00) - 2*u* (k_p00 - k_m00) + u*u * k_000 ;
		ex k_201 = (k_p01 + k_m01) - 2*u* (k_p01 - k_m01) + u*u * k_001 ;
		ex k_202 = (k_p02 + k_m02) - 2*u* (k_p02 - k_m02) + u*u * k_002 ;
		ex k_210 = (k_p10 + k_m10) - 2*u* (k_p10 - k_m10) + u*u * k_010 ;
		ex k_211 = (k_p11 + k_m11) - 2*u* (k_p11 - k_m11) + u*u * k_011 ;
		ex k_212 = (k_p12 + k_m12) - 2*u* (k_p12 - k_m12) + u*u * k_012 ;
		ex k_220 = (k_p20 + k_m20) - 2*u* (k_p20 - k_m20) + u*u * k_020 ;
		ex k_221 = (k_p21 + k_m21) - 2*u* (k_p21 - k_m21) + u*u * k_021 ;
		ex k_222 = (k_p22 + k_m22) - 2*u* (k_p22 - k_m22) + u*u * k_022 ;

		//Eq G2015(51)
		ex C_211 = k_211 - (k_200*k_011 + 2*k_101*k_110)/rho;
		ex C_121 = k_121 - (k_020*k_101 + 2*k_110*k_011)/rho;
		ex C_112 = k_112 - (k_002*k_110 + 2*k_011*k_101)/rho;

		//Eq G2015(52)
		ex C_220 = k_220- (k_020*k_200 + 2*k_110*k_110)/rho;
		ex C_022 = k_022- (k_002*k_020 + 2*k_011*k_011)/rho;
		ex C_202 = k_202- (k_200*k_002 + 2*k_101*k_101)/rho;

		//Eq G2015(53)
		ex C_122 = k_122- (k_020*k_102 + k_002*k_120 + 4*k_011*k_111 + 2*( k_110*k_012 + k_101*k_021))/rho;
		ex C_212 = k_212- (k_002*k_210 + k_200*k_012 + 4*k_101*k_111 + 2*( k_011*k_201 + k_110*k_102))/rho;
		ex C_221 = k_221- (k_200*k_021 + k_020*k_201 + 4*k_110*k_111 + 2*( k_101*k_120 + k_011*k_210))/rho;
		//Eq G2015(54)
		ex C_222 = k_222 - ( 4*k_111*k_111 + k_200*k_022 + k_020*k_202 + k_002*k_220 + 4*(k_011*k_211 + k_101*k_121 + k_110*k_112) + 2*(k_120*k_102 + k_210*k_012 + k_201*k_021))/rho +(16*k_110*k_101*k_011 + 4*(k_101*k_101*k_020 + k_011*k_011*k_200 + k_110*k_110*k_002)+2*k_200*k_020*k_002)/rho/rho;
		// gen1nowell.php END

		// 		// relaxation definition
		// 		ex omega1 = 1/(no3*KS.lbmViscosity+(ex)1/(ex)2); // shear viscosity
		// 		ex omega2 = 1;///(no3*KS.lbmViscosity*2 + (ex)1/(ex)2); // bulkViscosity > Viscosity ... test: bulkViscosity = 2/3 shearViscosity
		// 
		// 		ex omega3 = 1;
		// 		ex omega4 = 1;
		// 		ex omega5 = 1;
		// 		ex omega6 = 1;
		// 		ex omega7 = 1;
		// 		ex omega8 = 1;
		// 		ex omega9 = 1;
		// 		ex omega10 = 1;

		ex omega1 = sym_os[0];
		ex omega2 = sym_os[1];
		ex omega3 = sym_os[2];
		ex omega4 = sym_os[3];
		ex omega5 = sym_os[4];
		ex omega6 = sym_os[5];
		ex omega7 = sym_os[6];
		ex omega8 = sym_os[7];
		ex omega9 = sym_os[8];
		ex omega10 = sym_os[9];

		// extra parameters
		ex A=0;
		ex B=0;

		// actual collision step: note: ks = Cs for these indexes
		ex Cs_110 = (1 - omega1)*k_110;
		ex Cs_101 = (1 - omega1)*k_101;
		ex Cs_011 = (1 - omega1)*k_011;
		ex Dxu = 0;
		ex Dyv = 0;
		ex Dzw = 0;
		// plus their combination: Eq 30 - 32
		ex DxvDyu = 0;
		ex DxwDzu = 0;
		ex DywDzv = 0;

		// Eqs 33-35
		ex Eq33RHS = (1-omega1)*(k_200-k_020) - no3*rho*(1 - omega1*(ex)1/(ex)2)*(u*u*Dxu - v*v*Dyv);
		ex Eq34RHS = (1-omega1)*(k_200-k_002) - no3*rho*(1 - omega1*(ex)1/(ex)2)*(u*u*Dxu - w*w*Dzw);
		ex Eq35RHS = k_000*omega2*no3*sym_cs*sym_cs + (1-omega2)*(k_200+k_020+k_002) - no3*rho*(1-omega2/2)*(u*u*Dxu + v*v*Dyv + w*w*Dzw);
		// see rovnice2.mw
		ex Cs_200 = n1o3*(Eq33RHS + Eq34RHS + Eq35RHS);
		ex Cs_020 = n1o3*(-2*Eq33RHS + Eq34RHS + Eq35RHS);
		ex Cs_002 = n1o3*(Eq33RHS - 2*Eq34RHS + Eq35RHS);

		// Eqs 36-41: see vzorce2.mw
		ex Cs_120 = (-k_102-k_120)*omega3*(ex)1/(ex)2+(k_102-k_120)*omega4*(ex)1/(ex)2+k_120;
		ex Cs_102 = (-k_102-k_120)*omega3*(ex)1/(ex)2+(-k_102+k_120)*omega4*(ex)1/(ex)2+k_102;
		ex Cs_210 = (-k_012-k_210)*omega3*(ex)1/(ex)2+(k_012-k_210)*omega4*(ex)1/(ex)2+k_210;
		ex Cs_012 = (-k_012-k_210)*omega3*(ex)1/(ex)2+(-k_012+k_210)*omega4*(ex)1/(ex)2+k_012;
		ex Cs_021 = (-k_021-k_201)*omega3*(ex)1/(ex)2+(-k_021+k_201)*omega4*(ex)1/(ex)2+k_021;
		ex Cs_201 = (-k_021-k_201)*omega3*(ex)1/(ex)2+(k_021-k_201)*omega4*(ex)1/(ex)2+k_201;
		// Eq 42
		ex Cs_111 = (1-omega5)*k_111;

		// Eqs 43-45
		ex Eq43RHS = n2o3*(1/omega1-(ex)1/(ex)2)*omega6*A*rho*(Dxu - 2*Dyv + Dzw) + (1-omega6)*(C_220-2*C_202+C_022);
		ex Eq44RHS = n2o3*(1/omega1-(ex)1/(ex)2)*omega6*A*rho*(Dxu + Dyv -2* Dzw) + (1-omega6)*(C_220+C_202-2*C_022);
		ex Eq45RHS = -n4o3*(1/omega1-(ex)1/(ex)2)*omega7*A*rho*(Dxu + Dyv + Dzw) + (1-omega7)*(C_220+C_202+C_022);
		// see rovnice2.mw
		ex Cs_220 = n1o3*(Eq43RHS+Eq44RHS+Eq45RHS);
		ex Cs_202 = n1o3*(-Eq43RHS+Eq45RHS);
		ex Cs_022 = n1o3*(-Eq44RHS+Eq45RHS);
		// Eq 46-48
		ex Cs_211 = -n1o3*(1/omega1-(ex)1/(ex)2)*omega8*B*rho*DywDzv + (1-omega8)*C_211;
		ex Cs_121 = -n1o3*(1/omega1-(ex)1/(ex)2)*omega8*B*rho*DxwDzu + (1-omega8)*C_121;
		ex Cs_112 = -n1o3*(1/omega1-(ex)1/(ex)2)*omega8*B*rho*DxvDyu + (1-omega8)*C_112;
		// Eqs 49-52
		ex Cs_221 = (1-omega9)*C_221;
		ex Cs_212 = (1-omega9)*C_212;
		ex Cs_122 = (1-omega9)*C_122;
		ex Cs_222 = (1-omega10)*C_222;

		// remark: collision step is the same as in well-CUM

		// 3.4 Backward cumulant transformation
		// gen2nowell.php BEGIN
		//Eq. G2015(81)
		ex ks_211 = Cs_211+ (Cs_200*Cs_011 + 2*Cs_101*Cs_110)/rho;
		ex ks_121 = Cs_121+ (Cs_020*Cs_101 + 2*Cs_110*Cs_011)/rho;
		ex ks_112 = Cs_112+ (Cs_002*Cs_110 + 2*Cs_011*Cs_101)/rho;

		//Eq. G2015(82)
		ex ks_220 = Cs_220+ (Cs_020*Cs_200 + 2*Cs_110*Cs_110)/rho;
		ex ks_022 = Cs_022+ (Cs_002*Cs_020 + 2*Cs_011*Cs_011)/rho;
		ex ks_202 = Cs_202+ (Cs_200*Cs_002 + 2*Cs_101*Cs_101)/rho;

		//Eq. G2015(83)
		ex ks_122 = Cs_122+ (Cs_020*Cs_102 + Cs_002*Cs_120+ 4*Cs_011*Cs_111+ 2*(Cs_110*Cs_012+Cs_101*Cs_021))/rho;
		ex ks_212 = Cs_212+ (Cs_002*Cs_210 + Cs_200*Cs_012+ 4*Cs_101*Cs_111+ 2*(Cs_011*Cs_201+Cs_110*Cs_102))/rho;
		ex ks_221 = Cs_221+ (Cs_200*Cs_021 + Cs_020*Cs_201+ 4*Cs_110*Cs_111+ 2*(Cs_101*Cs_120+Cs_011*Cs_210))/rho;
		// gen2nowell.php END

		// Eq. G2015(84)
		ex ks_222 = Cs_222 + (4*Cs_111*Cs_111 + Cs_200*ks_022 + Cs_020*ks_202 + Cs_002*ks_220 + 4*(Cs_011*ks_211 + Cs_101*ks_121 + Cs_110*ks_112)
				+ 2*(Cs_120*Cs_102 + Cs_210*Cs_012 + Cs_201*Cs_021))/rho
				- (16*Cs_110*Cs_101*Cs_011 + 4*(Cs_101*Cs_101*Cs_020 + Cs_011*Cs_011*Cs_200 + Cs_110*Cs_110*Cs_002) + 2*Cs_200*Cs_020*Cs_002)/rho/rho;


		// backward central moment transformation
		ex ks_000 =  k_000;
		// Geier 2017: forcing scheme ---> zde se musi dat (-1) pro forcing ... ale s nulovou silou a - zde to nefunguje ... 
		ex Cs_100 =  -k_100;
		ex Cs_010 =  -k_010;
		ex Cs_001 =  -k_001;

		// gen3nowell.php BEGIN
		//Eq G2015(88)
		ex ks_z00 = ks_000*(1-u*u) - 2*u*Cs_100 - Cs_200;
		ex ks_z01 = Cs_001*(1-u*u) - 2*u*Cs_101 - Cs_201;
		ex ks_z02 = Cs_002*(1-u*u) - 2*u*Cs_102 - ks_202;
		ex ks_z10 = Cs_010*(1-u*u) - 2*u*Cs_110 - Cs_210;
		ex ks_z11 = Cs_011*(1-u*u) - 2*u*Cs_111 - ks_211;
		ex ks_z12 = Cs_012*(1-u*u) - 2*u*ks_112 - ks_212;
		ex ks_z20 = Cs_020*(1-u*u) - 2*u*Cs_120 - ks_220;
		ex ks_z21 = Cs_021*(1-u*u) - 2*u*ks_121 - ks_221;
		ex ks_z22 = ks_022*(1-u*u) - 2*u*ks_122 - ks_222;

		//Eq  G2015(89)
		ex ks_m00 = (ks_000*(u*u - u) + Cs_100*(2*u - 1) + Cs_200)*n1o2;
		ex ks_m01 = (Cs_001*(u*u - u) + Cs_101*(2*u - 1) + Cs_201)*n1o2;
		ex ks_m02 = (Cs_002*(u*u - u) + Cs_102*(2*u - 1) + ks_202)*n1o2;
		ex ks_m10 = (Cs_010*(u*u - u) + Cs_110*(2*u - 1) + Cs_210)*n1o2;
		ex ks_m11 = (Cs_011*(u*u - u) + Cs_111*(2*u - 1) + ks_211)*n1o2;
		ex ks_m12 = (Cs_012*(u*u - u) + ks_112*(2*u - 1) + ks_212)*n1o2;
		ex ks_m20 = (Cs_020*(u*u - u) + Cs_120*(2*u - 1) + ks_220)*n1o2;
		ex ks_m21 = (Cs_021*(u*u - u) + ks_121*(2*u - 1) + ks_221)*n1o2;
		ex ks_m22 = (ks_022*(u*u - u) + ks_122*(2*u - 1) + ks_222)*n1o2;

		//Eq  G2015(90)
		ex ks_p00 = (ks_000*(u*u + u) + Cs_100*(2*u + 1) + Cs_200)*n1o2;
		ex ks_p01 = (Cs_001*(u*u + u) + Cs_101*(2*u + 1) + Cs_201)*n1o2;
		ex ks_p02 = (Cs_002*(u*u + u) + Cs_102*(2*u + 1) + ks_202)*n1o2;
		ex ks_p10 = (Cs_010*(u*u + u) + Cs_110*(2*u + 1) + Cs_210)*n1o2;
		ex ks_p11 = (Cs_011*(u*u + u) + Cs_111*(2*u + 1) + ks_211)*n1o2;
		ex ks_p12 = (Cs_012*(u*u + u) + ks_112*(2*u + 1) + ks_212)*n1o2;
		ex ks_p20 = (Cs_020*(u*u + u) + Cs_120*(2*u + 1) + ks_220)*n1o2;
		ex ks_p21 = (Cs_021*(u*u + u) + ks_121*(2*u + 1) + ks_221)*n1o2;
		ex ks_p22 = (ks_022*(u*u + u) + ks_122*(2*u + 1) + ks_222)*n1o2;

		//Eq G2015(91)
		ex ks_mz0 = ks_m00*(1 - v*v) - 2*v*ks_m10 - ks_m20;
		ex ks_mz1 = ks_m01*(1 - v*v) - 2*v*ks_m11 - ks_m21;
		ex ks_mz2 = ks_m02*(1 - v*v) - 2*v*ks_m12 - ks_m22;
		ex ks_zz0 = ks_z00*(1 - v*v) - 2*v*ks_z10 - ks_z20;
		ex ks_zz1 = ks_z01*(1 - v*v) - 2*v*ks_z11 - ks_z21;
		ex ks_zz2 = ks_z02*(1 - v*v) - 2*v*ks_z12 - ks_z22;
		ex ks_pz0 = ks_p00*(1 - v*v) - 2*v*ks_p10 - ks_p20;
		ex ks_pz1 = ks_p01*(1 - v*v) - 2*v*ks_p11 - ks_p21;
		ex ks_pz2 = ks_p02*(1 - v*v) - 2*v*ks_p12 - ks_p22;

		//Eq  G2015(92)
		ex ks_mm0 = (ks_m00*(v*v - v) + ks_m10*(2*v - 1) + ks_m20)*n1o2;
		ex ks_mm1 = (ks_m01*(v*v - v) + ks_m11*(2*v - 1) + ks_m21)*n1o2;
		ex ks_mm2 = (ks_m02*(v*v - v) + ks_m12*(2*v - 1) + ks_m22)*n1o2;
		ex ks_zm0 = (ks_z00*(v*v - v) + ks_z10*(2*v - 1) + ks_z20)*n1o2;
		ex ks_zm1 = (ks_z01*(v*v - v) + ks_z11*(2*v - 1) + ks_z21)*n1o2;
		ex ks_zm2 = (ks_z02*(v*v - v) + ks_z12*(2*v - 1) + ks_z22)*n1o2;
		ex ks_pm0 = (ks_p00*(v*v - v) + ks_p10*(2*v - 1) + ks_p20)*n1o2;
		ex ks_pm1 = (ks_p01*(v*v - v) + ks_p11*(2*v - 1) + ks_p21)*n1o2;
		ex ks_pm2 = (ks_p02*(v*v - v) + ks_p12*(2*v - 1) + ks_p22)*n1o2;

		//Eq G2015(93)
		ex ks_mp0 = (ks_m00*(v*v + v) + ks_m10*(2*v + 1) + ks_m20)*n1o2;
		ex ks_mp1 = (ks_m01*(v*v + v) + ks_m11*(2*v + 1) + ks_m21)*n1o2;
		ex ks_mp2 = (ks_m02*(v*v + v) + ks_m12*(2*v + 1) + ks_m22)*n1o2;
		ex ks_zp0 = (ks_z00*(v*v + v) + ks_z10*(2*v + 1) + ks_z20)*n1o2;
		ex ks_zp1 = (ks_z01*(v*v + v) + ks_z11*(2*v + 1) + ks_z21)*n1o2;
		ex ks_zp2 = (ks_z02*(v*v + v) + ks_z12*(2*v + 1) + ks_z22)*n1o2;
		ex ks_pp0 = (ks_p00*(v*v + v) + ks_p10*(2*v + 1) + ks_p20)*n1o2;
		ex ks_pp1 = (ks_p01*(v*v + v) + ks_p11*(2*v + 1) + ks_p21)*n1o2;
		ex ks_pp2 = (ks_p02*(v*v + v) + ks_p12*(2*v + 1) + ks_p22)*n1o2;

		//Eq G2015(94)
		rhs(mmz, 0) =  ks_mm0*(1-w*w) - 2*w*ks_mm1 - ks_mm2;
		rhs(mzz, 0) =  ks_mz0*(1-w*w) - 2*w*ks_mz1 - ks_mz2;
		rhs(mpz, 0) =  ks_mp0*(1-w*w) - 2*w*ks_mp1 - ks_mp2;
		rhs(zmz, 0) =  ks_zm0*(1-w*w) - 2*w*ks_zm1 - ks_zm2;
		rhs(zzz, 0) =  ks_zz0*(1-w*w) - 2*w*ks_zz1 - ks_zz2;
		rhs(zpz, 0) =  ks_zp0*(1-w*w) - 2*w*ks_zp1 - ks_zp2;
		rhs(pmz, 0) =  ks_pm0*(1-w*w) - 2*w*ks_pm1 - ks_pm2;
		rhs(pzz, 0) =  ks_pz0*(1-w*w) - 2*w*ks_pz1 - ks_pz2;
		rhs(ppz, 0) =  ks_pp0*(1-w*w) - 2*w*ks_pp1 - ks_pp2;

		//Eq  G2015(95)
		rhs(mmm, 0) =  (ks_mm0*(w*w-w) + ks_mm1*(2*w - 1) + ks_mm2)*n1o2;
		rhs(mzm, 0) =  (ks_mz0*(w*w-w) + ks_mz1*(2*w - 1) + ks_mz2)*n1o2;
		rhs(mpm, 0) =  (ks_mp0*(w*w-w) + ks_mp1*(2*w - 1) + ks_mp2)*n1o2;
		rhs(zmm, 0) =  (ks_zm0*(w*w-w) + ks_zm1*(2*w - 1) + ks_zm2)*n1o2;
		rhs(zzm, 0) =  (ks_zz0*(w*w-w) + ks_zz1*(2*w - 1) + ks_zz2)*n1o2;
		rhs(zpm, 0) =  (ks_zp0*(w*w-w) + ks_zp1*(2*w - 1) + ks_zp2)*n1o2;
		rhs(pmm, 0) =  (ks_pm0*(w*w-w) + ks_pm1*(2*w - 1) + ks_pm2)*n1o2;
		rhs(pzm, 0) =  (ks_pz0*(w*w-w) + ks_pz1*(2*w - 1) + ks_pz2)*n1o2;
		rhs(ppm, 0) =  (ks_pp0*(w*w-w) + ks_pp1*(2*w - 1) + ks_pp2)*n1o2;

		//Eq  G2015(96)
		rhs(mmp, 0) =  (ks_mm0*(w*w+w) + ks_mm1*(2*w + 1) + ks_mm2)*n1o2;
		rhs(mzp, 0) =  (ks_mz0*(w*w+w) + ks_mz1*(2*w + 1) + ks_mz2)*n1o2;
		rhs(mpp, 0) =  (ks_mp0*(w*w+w) + ks_mp1*(2*w + 1) + ks_mp2)*n1o2;
		rhs(zmp, 0) =  (ks_zm0*(w*w+w) + ks_zm1*(2*w + 1) + ks_zm2)*n1o2;
		rhs(zzp, 0) =  (ks_zz0*(w*w+w) + ks_zz1*(2*w + 1) + ks_zz2)*n1o2;
		rhs(zpp, 0) =  (ks_zp0*(w*w+w) + ks_zp1*(2*w + 1) + ks_zp2)*n1o2;
		rhs(pmp, 0) =  (ks_pm0*(w*w+w) + ks_pm1*(2*w + 1) + ks_pm2)*n1o2;
		rhs(pzp, 0) =  (ks_pz0*(w*w+w) + ks_pz1*(2*w + 1) + ks_pz2)*n1o2;
		rhs(ppp, 0) =  (ks_pp0*(w*w+w) + ks_pp1*(2*w + 1) + ks_pp2)*n1o2;
	}
};
